using System;
using System.Collections.Generic;
using System.Text;

namespace Unit4.CollectionsLib
{
    /// <summary>
    /// <h3 dir="rtl" style="color:red; font-family: Helvetica, Arial, sans-serif">
    ///        .       
    /// </h3>
    /// </summary>
    /// <author>   ,   ,  ,  </author>
    /// <version> 26.11.2007 </version>
    /// <typeparam name="V">     </typeparam>
    public class Map<V>
    {
        private BinTreeNode<Pair<string, V>> root;
        private int numOfKeys;

        /// <summary>
        /// <dt dir="rtl"/>
        /// <b>
        ///     
        /// </b>
        /// </summary>
        public Map()
        {
            this.root = null;
            this.numOfKeys = 0;
        }

        /// <summary>
        ///   ''   key    -'' 
        /// </summary>
        /// <param name="key">  </param>
        /// <returns>   ''    , ''  </returns>
        public bool ContainsKey(string key)
        {
            BinTreeNode<Pair<string, V>> pos = SearchKey(key);

            if (pos == null)
                return false;
            return true;
        }

        /// <summary>
        /// <dt dir="rtl"/>
        /// <b>
        ///       key
        /// <br/>
        /// <u>:</u>   
        /// </b>
        /// </summary>
        /// <param name="key">  </param>
        /// <returns>   </returns>
        public V GetValue(string key)
        {
            BinTreeNode<Pair<string, V>> pos = SearchKey(key);
            return pos.GetInfo().GetValue();
        }

        /// <summary>
        /// <dt dir="rtl"/>
        /// <b>
        ///      key   value  ;
        ///   key  ,        value   
        /// </b>
        /// </summary> 
        /// <param name="key">   </param>
        /// <returns>   </returns>
        public void Insert(string key, V value)
        {
            BinTreeNode<Pair<string, V>> pos = SearchKey(key);

            if (pos != null)
                pos.GetInfo().SetValue(value);
            else
            {
                this.numOfKeys++;
                pos = this.root;
                bool inTree = false;
                BinTreeNode<Pair<string, V>> node = new BinTreeNode<Pair<string, V>>(new Pair<string, V>(key, value));

                if (this.root == null)
                {
                    this.root = node;
                    inTree = true;
                }
                while (!inTree)
                {
                    if (key.CompareTo(pos.GetInfo().GetKey()) < 0)
                    {
                        if (pos.GetLeft() == null)
                        {
                            pos.SetLeft(node);
                            inTree = true;
                        }
                        else
                            pos = pos.GetLeft();
                    }
                    else
                    {
                        if (pos.GetRight() == null)
                        {
                            pos.SetRight(node);
                            inTree = true;
                        }
                        else
                            pos = pos.GetRight();
                    }
                }
            }
        }

        /// <summary>
        /// <dt dir="rtl"/>
        /// <b>
        ///       key    ;
        ///        .
        /// <br/>
        /// <u>:</u>   
        /// </b>
        /// </summary>
        /// <param name="key">   </param>
        /// <returns>    </returns>
        public V Remove(string key)
        {
            BinTreeNode<Pair<string, V>> pos = SearchKey(key);

            this.numOfKeys--;

            BinTreeNode<Pair<string, V>> next = null;
            BinTreeNode<Pair<string, V>> parent = GetParent(pos);

            if ((pos.GetLeft() == null) || (pos.GetRight() == null))
            {
                if (pos.GetLeft() != null)
                {
                    next = pos.GetLeft();
                    pos.SetLeft(null);
                }
                else
                {
                    next = pos.GetRight();
                    pos.SetRight(null);
                }
                if (parent == null)
                    this.root = next;
                else
                    if (parent.GetLeft() == pos)
                        parent.SetLeft(next);
                    else
                        parent.SetRight(next);
            }
            else
            {
                next = pos.GetRight();
                while (next.GetLeft() != null)
                    next = next.GetLeft();
                pos.SetInfo(next.GetInfo());
                if (pos.GetRight() != next)
                {
                    parent = GetParent(next);
                    parent.SetLeft(next.GetRight());
                }
                else
                    pos.SetRight(next.GetRight());
                next.SetRight(null);
            }

            return pos.GetInfo().GetValue();
        }

        /// <summary>
        /// <dt dir="rtl"/>
        /// <b>
        ///            .
        ///   ,    	
        /// </b>
        /// </summary> 
        /// <returns>          </returns>
        public string[] GetAllKeys()
        {
            int i = 0;
            BinTreeNode<Pair<string, V>> tmp = this.root;
            string[] keys = new string[this.numOfKeys];
            Stack<BinTreeNode<Pair<string, V>>> stk = new Stack<BinTreeNode<Pair<string, V>>>();

            do
            {
                while (tmp != null)
                {
                    stk.Push(tmp);
                    tmp = tmp.GetLeft();
                }
                if (!stk.IsEmpty())
                {
                    tmp = stk.Pop();
                    keys[i] = tmp.GetInfo().GetKey();
                    i++;
                    tmp = tmp.GetRight();
                }
            } while (!stk.IsEmpty() || tmp != null);
            return (keys);
        }

        /// <summary>
        /// <dt dir="rtl"/>
        /// <b>
        ///      
        /// </b>
        /// </summary>
        /// <returns> [key1:value1, key2:value2, key3:value3, ...] :      </returns>
        public override string ToString()
        {
            int i;
            string str = "[";
            string[] keys = this.GetAllKeys();

            for (i = 0; i < keys.Length - 1; i++)
                str = str + keys[i] + ":" + this.GetValue(keys[i]).ToString() + ", ";
            if (keys.Length > 0)
                str = str + keys[i] + ":" + this.GetValue(keys[i]).ToString();
            str = str + "]";
            return (str);
        }


        // ------------------ private methods for help -----------------
        private BinTreeNode<Pair<string, V>> SearchKey(string key)
        {
            BinTreeNode<Pair<string, V>> p = this.root;

            while (p != null)
            {
                if (key.CompareTo(p.GetInfo().GetKey()) == 0)
                    return (p);
                if (key.CompareTo(p.GetInfo().GetKey()) < 0)
                    p = p.GetLeft();
                else
                    p = p.GetRight();
            }
            return null;
        }

        private BinTreeNode<Pair<string, V>> GetParent(BinTreeNode<Pair<string, V>> node)
        {
            BinTreeNode<Pair<string, V>> p = this.root;

            while (p != null)
            {
                if ((node == p.GetLeft()) || (node == p.GetRight()))
                    return (p);
                if (node.GetInfo().GetKey().CompareTo(p.GetInfo().GetKey()) >= 0)
                    p = p.GetRight();
                else
                    p = p.GetLeft();
            }
            return (null);
        }
    }
}
