using System;
//using System.Collections.Generic;
using System.Text;
using Unit4.CollectionsLib;

namespace Unit4.UtilsLib
{
    /// <summary>
    /// <h3 dir="rtl" style="color:red; font-family: Helvetica, Arial, sans-serif">
    ///            
    /// </h3>
    /// </summary>
    /// <author>   ,   ,  ,  </author>
    /// <version> 27.11.2007 </version>
    public class BinTreeUtils
    {
        private BinTreeUtils() { }
        private static Random random = new Random();
        
        /// <summary>
        /// <dt dir="rtl"/>
        /// <b>
        ///          
        /// </b>
        /// </summary>
        /// <param name="maxNodes">    </param>
        /// <param name="low">      </param>
        /// <param name="high">      </param>
        /// <returns>      </returns>
        public static BinTreeNode<int> BuildRandomTree(int maxNodes, int low, int high)
        {
            int dir = 0;
            BinTreeNode<int> t1, t2 = null;
            int x = BinTreeUtils.random.Next(low, high + 1);
            BinTreeNode<int> tree = new BinTreeNode<int>(x);

            for (int i = 2; i <= maxNodes; i++)
            {
                t1 = tree;
                while (t1 != null)
                {
                    t2 = t1;
                    dir = (int)(BinTreeUtils.random.NextDouble() * 2);
                    if (dir == 1)
                        t1 = t1.GetLeft();
                    else
                        t1 = t1.GetRight();
                }
                x = BinTreeUtils.random.Next(low, high + 1);
                if (dir == 1)
                    t2.SetLeft(new BinTreeNode<int>(x));
                else
                    t2.SetRight(new BinTreeNode<int>(x));
            }

            return (tree);
        }

        //----------------------- traversals --------------------------
       
        /// <summary>
        /// <dt dir="rtl"/>
        /// <b>
        ///         
        /// </b>
        /// </summary>
        /// <param name="tree">    </param>
        /// <returns>        </returns>
        public static string PreOrderTraversal<T>(BinTreeNode<T> tree)
        {
            string str = "";

            str = str + tree.GetInfo() + " , ";

            if (tree.GetLeft() != null)
                str = str + PreOrderTraversal(tree.GetLeft());

            if (tree.GetRight() != null)
                str = str + PreOrderTraversal(tree.GetRight());

            return (str);
        }

        /// <summary>
        /// <dt dir="rtl"/>
        /// <b>
        ///         
        /// </b>
        /// </summary>
        /// <param name="tree">    </param>
        /// <returns>        </returns>
        public static string InOrderTraversal<T>(BinTreeNode<T> tree)
        {
            string str = "";

            if (tree.GetLeft() != null)
                str = str + InOrderTraversal(tree.GetLeft());

            str = str + tree.GetInfo() + " , ";

            if (tree.GetRight() != null)
                str = str + InOrderTraversal(tree.GetRight());

            return (str);
        }

        /// <summary>
        /// <dt dir="rtl"/>
        /// <b>
        ///         
        /// </b>
        /// </summary>
        /// <param name="tree">    </param>
        /// <returns>        </returns>
        public static string PostOrderTraversal<T>(BinTreeNode<T> tree)
        {
            string str = "";

            if (tree.GetLeft() != null)
                str = str + PostOrderTraversal(tree.GetLeft());

            if (tree.GetRight() != null)
                str = str + PostOrderTraversal(tree.GetRight());

            str = str + tree.GetInfo() + " , ";

            return (str);
        }

          
        /// <summary>
        /// <dt dir="rtl"/>
        /// <b>
        ///         
        /// </b>
        /// </summary>
        /// <param name="tree">   </param>
        /// <returns>        </returns>
        public static string LevelOrderTraversal<T>(BinTreeNode<T> tree)
        {
            BinTreeNode<T> t;
            string str = "";
            Queue<BinTreeNode<T>> q = new Queue<BinTreeNode<T>>();

            q.Insert(tree);
            while (!q.IsEmpty())
            {
                t = q.Remove();
                str = str + t.GetInfo() + " , ";
                if (t.GetLeft() != null)
                    q.Insert(t.GetLeft());
                if (t.GetRight() != null)
                    q.Insert(t.GetRight());
            }

            return (str);
        }

        /// <summary>
        /// <dt dir="rtl"/>
        /// <b>
        ///      (   0)
        /// </b>
        /// </summary>
        /// <param name="tree">    </param>
        /// <returns>   </returns>
        public static int Height<T>(BinTreeNode<T> tree)
        {
            int hl = 0;
            int hr = 0;

            if (tree.GetLeft() != null)
                hl = 1 + Height(tree.GetLeft());
            if (tree.GetRight() != null)
                hr = 1 + Height(tree.GetRight());
            return (Math.Max(hl, hr));
        }

        /// <summary>
        /// <dt dir="rtl"/>
        /// <b>
        ///      
        /// </b>
        /// </summary>
        /// <param name="tree">    </param>
        /// <returns>    </returns>
        public static int NumOfNodes<T>(BinTreeNode<T> tree)
        {
            int nl = 0;
            int nr = 0;

            if (tree.GetLeft() != null)
                nl = NumOfNodes(tree.GetLeft());
            if (tree.GetRight() != null)
                nr = NumOfNodes(tree.GetRight());
            return (1 + nl + nr);
        }

        /// <summary>
        /// <dt dir="rtl"/>
        /// <b>
        ///         ()   
        /// </b>
        /// </summary>
        /// <param name="tree">    </param>
        /// <returns>   </returns>
        public static BinTreeNode<T> Clone<T>(BinTreeNode<T> tree)
        {
            BinTreeNode<T> left = null;
            BinTreeNode<T> right = null;

            if (tree.GetLeft() != null)
                left = Clone(tree.GetLeft());
            if (tree.GetRight() != null)
                right = Clone(tree.GetRight());

            return (new BinTreeNode<T>(left, tree.GetInfo(), right));
        }

        /// <summary>
        /// <dt dir="rtl"/>
        /// <b>
        ///     
        /// </b>
        /// </summary>
        /// <param name="tree">    </param>
        /// <returns>   ''   , ''  </returns>
        public static bool IsFull<T>(BinTreeNode<T> tree)
        {
            if (BinTreeUtils.IsLeaf(tree))
                return (true);

            if (tree.GetLeft() == null || tree.GetRight() == null)
                return (false);

            if (BinTreeUtils.Height(tree.GetLeft()) != BinTreeUtils.Height(tree.GetRight()))
                return (false);

            return (IsFull(tree.GetLeft()) && IsFull(tree.GetRight()));
        }

        /// <summary>
        /// <dt dir="rtl"/>
        /// <b>
        ///           
        /// </b>
        /// </summary>
        /// <param name="tree">      </param>
        /// <returns>     </returns>
        public static int Max(BinTreeNode<int> tree)
        {
            int ml = tree.GetInfo();
            int mr = tree.GetInfo();

            if (tree.GetLeft() != null)
                ml = Math.Max(tree.GetInfo(), Max(tree.GetLeft()));
            if (tree.GetRight() != null)
                mr = Math.Max(tree.GetInfo(), Max(tree.GetRight()));
            return (Math.Max(ml, mr));
        }

        /// <summary>
        /// <dt dir="rtl"/>
        /// <b>
        ///          
        /// </b>
        /// </summary>
        /// <param name="tree">      </param>
        /// <returns>      </returns>
        public static int SumOfNodes(BinTreeNode<int> tree)
        {
            int sl = 0;
            int sr = 0;

            if (tree.GetLeft() != null)
                sl = SumOfNodes(tree.GetLeft());
            if (tree.GetRight() != null)
                sr = SumOfNodes(tree.GetRight());
            return (tree.GetInfo() + sl + sr);
        }

        /// <summary>
        /// <dt dir="rtl"/>
        /// <b>
        ///            
        /// </b>
        /// </summary>
        /// <param name="tree">      </param>
        /// <param name="level">  (    ) </param>
        /// <returns>       </returns>
        public static int SumOfNodesInLevel(BinTreeNode<int> tree, int level)
        {
            int sl = 0;
            int sr = 0;

            if (level == 0)
                return (tree.GetInfo());
            if (tree.GetLeft() != null)
                sl = SumOfNodesInLevel(tree.GetLeft(), level - 1);
            if (tree.GetRight() != null)
                sr = SumOfNodesInLevel(tree.GetRight(), level - 1);
            return (sl + sr);
        }

        /// <summary>
        /// <dt dir="rtl"/>
        /// <b>
        ///       
        /// </b>
        /// </summary>
        /// <param name="tree">      </param>
        /// <returns>  ''     , ''  </returns>
        public static bool IsAllPositive(BinTreeNode<int> tree)
        {
            bool resL = true;
            bool resR = true;

            if (tree.GetInfo() < 0)
                return (false);
            if (tree.GetLeft() != null)
                resL = IsAllPositive(tree.GetLeft());
            if (tree.GetRight() != null)
                resR = IsAllPositive(tree.GetRight());
            return (resL && resR);
        }

        /// <summary>
        /// <dt dir="rtl"/>
        /// <b>
        ///       
        /// </b>
        /// </summary>
        /// <param name="tree">      </param>
        /// <param name="x">   </param>
        /// <returns>   ''     , ''  </returns>
        public static bool Exists(BinTreeNode<int> tree, int x)
        {
            bool resL = false;
            bool resR = false;

            if (tree.GetInfo() == x)
                return (true);
            if (tree.GetLeft() != null)
                resL = Exists(tree.GetLeft(), x);
            if (tree.GetRight() != null)
                resR = Exists(tree.GetRight(), x);
            return (resL || resR);
        }
       
        /// <summary>
        /// <dt dir="rtl"/>
        /// <b>
        ///        
        /// </b>
        /// </summary>
        /// <param name="tree">    </param>
        /// <returns>   ''     , ''  </returns>
        public static bool IsLeaf<T>(BinTreeNode<T> tree)
        {
            return ((tree.GetLeft() == null) && (tree.GetRight() == null));
        }

        /// <summary>
        /// <dt dir="rtl"/>
        /// <b>
        ///      
        /// </b>
        /// </summary>
        /// <param name="tree">    </param>
        /// <returns>    </returns>
        public static int NumOfLeaves<T>(BinTreeNode<T> tree)
        {
            int nl = 0;
            int nr = 0;

            if (tree.GetLeft() == null && tree.GetRight() == null)
                return (1);
            if (tree.GetLeft() != null)
                nl = NumOfLeaves(tree.GetLeft());
            if (tree.GetRight() != null)
                nr = NumOfLeaves(tree.GetRight());
            return (nl + nr);
        }

        /// <summary>
        /// <dt dir="rtl"/>
        /// <b>
        ///         ;       null 
        /// </b>
        /// </summary>
        /// <param name="tree">    </param>
        /// <param name="child">    </param>
        /// <returns>      </returns>
        public static BinTreeNode<T> Parent<T>(BinTreeNode<T> tree, BinTreeNode<T> child)
        {
            if ((tree == null) || (child == tree.GetLeft()) || (child == tree.GetRight()))
                return (tree);
            BinTreeNode<T> tmp = Parent(tree.GetLeft(), child);
            if (tmp == null)
                return (Parent(tree.GetRight(), child));
            else
                return (tmp);
        }
    }
}
